/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *  
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License. 
 * 
 * The Original Code is pilot-mailsync.
 * 
 * The Initial Developer of the Original Code is Hollis Blanchard
 * <hollis@austin.rr.com>. Portions created by Hollis Blanchard are Copyright
 * (C) 2001 Hollis Blanchard. All Rights Reserved.
 * 
 * Contributor(s):
 * Frederic Devernay <Frederic.Devernay@sophia.inria.fr>
 * Stephan Allner <allner@informatik.hu-berlin.de>
 * Gordon Heydon <gjheydon@bigfoot.com>
 * Damien Burke <damienb@mit.edu>
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "c-client/c-client.h"

#include "cclientcalls.h"
#include "options.h"
#include "pilot-mailsync.h"
#include "llist.h"
#include "util.h"

#ifdef DEBUG
static void printMail(const struct Mail *mail);
#endif

static char *getHeader(MAILSTREAM *stream, unsigned long msg_no, const char *headerstr);
static void getMsg(MAILSTREAM *stream, unsigned long number);

static MAILSTREAM *gMailStream;
/* everything that we got back from a search */
static listItem *gSearchResults;

/* helper functions to create/destroy c-client STRINGLIST's */
static STRINGLIST *
mkSL(char *str)
{
	STRINGLIST *sl = (STRINGLIST *)malloc(sizeof(STRINGLIST));
	sl->text.data = strdup(str);
	sl->text.size = strlen(str);
	sl->next = NULL;
	return sl;
}

void
delSL(STRINGLIST *sl)
{
	if (sl == NULL)
		return;
	free(sl->text.data);
	free(sl);
}



/* SMTPMail()
 *
 * send mail using stmp, called by PalmToMail() (pilot-mailsync.c)
 */
int
SMTPMail(struct Mail *mail, const char *from, const char *charset)
{
	/* XXX implement me */
	return 0;
}

/* MailToString()
 *
 * converts a PalmMail to a (char *) with CRLF instead of LF ( :( )
 *
 * caller is responsible for freeing memory.
 */
char *
MailToString(char *mailstring, const struct Mail *mail, const char *from,
				const char *date, const char *charset)
{
	char *line=NULL, *crlf_string=NULL;
	int i;

	/* mailtab represents each line of the mail header. each non-empty line
	 * will be added to the mail.
	 */
	const char *mailtab[][2] = 	{
		{ "From: %s\n", from },
		{ "To: %s\n", mail->to },
		{ "Cc: %s\n", mail->cc },
		{ "Bcc: %s\n", mail->bcc },
		{ "Reply-To: %s\n", mail->replyTo },
		{ "Subject: %s\n", mail->subject ? mail->subject : "<none>" },
		{ "X-Mailer: %s\n", PACKAGE "-" VERSION},
		{ "Content-Type: text/plain; charset=%s\n", charset},
		{ "Date: %s\n", date },
		{ "\n", "" },
		{ "%s\n", mail->body },
		{ "-- \n%s\n", mail->signature ? gSig.signature : "" }
	};

	for (i=0; i<12; i++) {
		if (mailtab[i][1] && strlen(mailtab[i][1])) {
			/* make a buffer for this line */
			line = (char *)realloc(line, strlen(mailtab[i][0]) +
										strlen(mailtab[i][1]) + 1);
			sprintf(line, mailtab[i][0], mailtab[i][1]);

			/* resize the mailstring to fit the new line */
			mailstring = (char *)realloc(mailstring, strlen(mailstring) +
											strlen(line)+1);

			/* append this line to the string */
			strcat(mailstring, line);
		}
	}
	free(line);

	/* take our hard-earned string and make sure it contains only CRLF's */
	lf2crlf(&crlf_string, mailstring);
	mailstring = (char *)realloc(mailstring, strlen(crlf_string)+1);
	strncpy(mailstring, crlf_string, strlen(crlf_string)+1);
	free(crlf_string);

	return mailstring;
}

/* AddMailToMailbox()
 *
 * Adds a PalmMail to the specified mailbox
 */
int
AddMailToMailbox(char *mailbox, struct Mail *mail, const char *from) {
	STRING msg;
	char *mailstring;
	char date[MAILTMPLEN], flags[MAILTMPLEN];

	rfc822_date(date);
	strcpy(flags, "(\\Seen)"); /* mark mail as read */

	mailstring = (char *)malloc(1);
	mailstring[0] = '\0';
	mailstring = MailToString(mailstring, mail, from, date, gCharSet);

	/* write out the standard email headers */
	INIT(&msg, mail_string, (void *)mailstring, strlen(mailstring));
	mail_append_full(NIL, mailbox, flags, date, &msg);

	free(mailstring);
	return 0;
}

/* OpenMail()
 * 
 * opens connection to the remote mailbox
 */
void
OpenMail(char *mailbox)
{
	MAILSTREAM *stream = NIL;
	long options = 0;

/********* include a .c file... **********/
#include "c-client/linkage.c"
/*****************************************/

#ifdef DEBUG
	options = OP_DEBUG;
#endif

	mail_parameters(stream, SET_RSHTIMEOUT, 0); /* disable RSH attempts */
	if (NULL == (stream = mail_open(stream, mailbox, options))) {
		mm_log("fatal error: couldn't open mailbox", ERROR);
		return;
	}

	gMailStream = stream;
}

/* CloseMail()
 *
 * closes the mail connection
 */
void
CloseMail(void)
{
	mail_close_full(gMailStream, CL_EXPUNGE); /* expunge deleted mails */
}


/* MailToPalm()
 *
 * transfer mail from desktop to Palm
 */
void
MailToPalm(void)
{
	SEARCHPGM search, notsearch, imptsearch, moresearch;
	SEARCHPGM *curpgm=&search;
	SEARCHPGMLIST notlist;
	SEARCHOR orsearch;
	int msgno;

	memset(&search, 0, sizeof(SEARCHPGM));
	memset(&imptsearch, 0, sizeof(SEARCHPGM)); /* get flagged msgs */
	memset(&moresearch, 0, sizeof(SEARCHPGM)); /* complement to imptsearch */
	memset(&notsearch, 0, sizeof(SEARCHPGM));

	memset(&orsearch, 0, sizeof(SEARCHOR));
	memset(&notlist, 0, sizeof(SEARCHPGMLIST));

	if (gPrefs.syncType == mailSyncUnread) {
		if (gSyncSeen) {
			/* if we are updating the seen flag later, we need to only get
			 * recent messages (otherwise we'll get duplicates)
			 */
			search.recent = 1;
		} else {
			/* search for new messages */
			search.unseen = 1;
		}

	} else if (gPrefs.syncType == mailSyncAll){
		/* search pgm is already set to match everything */

		int len, size, attr;
		char buffer[0xffff];
		recordid_t id;

		/* delete all the emails in the inbox and replace them. If the PalmDB
		 * had a Message-ID field we could probably avoid this and only
		 * download the messages that aren't on the Palm already. DNB
		 */
		while (0 <= (len = dlp_ReadNextRecInCategory(gSocket, gDB,
								mailCtgInbox, buffer,
								&id, 0, &size, &attr))){
			printf("D"); fflush(stdout); /* DEBUG OUTOUT */
			dlp_DeleteRecord(gSocket, gDB, 0, id);
		}
	} else if (gPrefs.syncType == mailSyncFilter) {
		/* set up some filters */

		if (gPrefs.getHigh) {
			/* get all Important msgs, regardless of include/exclude */
			search.or = &orsearch;
			orsearch.first = &imptsearch; /* Important messages OR ... */
			orsearch.second = &moresearch;
			imptsearch.flagged = 1;
			curpgm = &moresearch; /* future filter settings go to or.second */
		}

		if (gPrefs.getContaining) {
			/* "include" search */
			if (gPrefs.filterTo)
				curpgm->to = mkSL(gPrefs.filterTo);
			if (gPrefs.filterFrom)
				curpgm->from = mkSL(gPrefs.filterFrom);
			if (gPrefs.filterSubject)
				curpgm->subject = mkSL(gPrefs.filterSubject);
		} else {
			/* "exclude" search */
			curpgm->not = &notlist;
			notlist.pgm = &notsearch;

			if (gPrefs.filterTo)
				notsearch.to = mkSL(gPrefs.filterTo);
			if (gPrefs.filterFrom)
				notsearch.from = mkSL(gPrefs.filterFrom);
			if (gPrefs.filterSubject)
				notsearch.subject = mkSL(gPrefs.filterSubject);
		}
	}

	/* this stores all messages in gSearchResults via mm_searched() */
	mail_search_full(gMailStream, gCharSet, &search, NIL);

	/* now process the messages that matched */
	while ((msgno = (int)pop(&gSearchResults)))
		getMsg(gMailStream, msgno);
	printf("\n"); /* terminate "...." from getMsg */

	/* free mkSL memory */
	delSL(curpgm->to);
	delSL(curpgm->from);
	delSL(curpgm->subject);
	delSL(notsearch.to);
	delSL(notsearch.from);
	delSL(notsearch.subject);
}

/* getMsg
 *
 * do the actual message processing:
 * - get from c-client (delete if appropriate)
 * - construct new Palm Mail item
 * - transfer to Palm
 */
void
getMsg(MAILSTREAM *stream, unsigned long number)
{
	BODY *body;
	MESSAGECACHE *elt;

	mail_fetchstructure(stream, number, &body);

	if (body) {
		/* 8000 bytes is the max truncation size, plus 500 bytes for to, cc,
		 * from, and subject lines
		 */
		unsigned char record[8000 + 500];
		struct Mail pilotMail;
		char *datestr;
		unsigned long len=0;

		/* mark mail to be deleted */
		if (gDelete && (gPrefs.syncType != mailSyncUnread)) {
			/* Palm Mail app says "all messages will remain on the E-Mail
			 * application" in sync mode Unread.
			 */
			markDeleted(number);
		}

		/* I don't know what the following do for Palm-bound mail */
		pilotMail.signature = 0;
		pilotMail.confirmRead = 0;
		pilotMail.confirmDelivery = 0;
		pilotMail.addressing = 0;

		elt = mail_elt(stream, number); /* for msg flags */

		/* in the Palm OS Mail struct, 0 is high priority (presumably 2 is low,
		 * but low doesn't seem to be visibly marked in the Mail app)
		 */
		pilotMail.priority = !(elt->flagged);
		pilotMail.read = elt->seen;

		/* convert message's text-date to struct tm */
		datestr = strip_cr(getHeader(stream, number, "Date"));
		pilotMail.date = *DateToTm(datestr);
		pilotMail.dated = 1;

		/* fill in all the text headers */
		pilotMail.subject = strip_cr(getHeader(stream, number, "Subject"));
		pilotMail.from = strip_cr(getHeader(stream, number, "From"));
		pilotMail.to = strip_cr(getHeader(stream, number, "To"));
		pilotMail.cc = strip_cr(getHeader(stream, number, "Cc"));
		pilotMail.bcc = strip_cr(getHeader(stream, number, "Bcc"));
		pilotMail.replyTo = strip_cr(getHeader(stream, number, "Reply-To"));
		pilotMail.sentTo = strip_cr(getHeader(stream, number, "Sent-To"));

		/* get body, truncating if necessary. Note that fetchtext_full
		 * should return the entire body without seperating parts
		 */
		pilotMail.body = strip_cr(
							mail_fetchtext_full(stream, number,
												&len,
												gSyncSeen ? FT_PEEK : NIL)
								);
		if (len > gPrefs.truncate) {
			/* write over the last few chars of the mail */
			char *trunc_warning = "\n\n[pilot-mailsync truncated]";
			strcpy(pilotMail.body+gPrefs.truncate-strlen(trunc_warning),
					trunc_warning);
			printf("/");
			fflush(stdout);
			if (gVerbose)
				printf("\n  truncating mail #%li; new len is %i\n", number,
						strlen(pilotMail.body));
#ifdef DEBUG
			printMail(&pilotMail);
#endif
		}

		/* finished setting all of the headers; write it to the record buffer */
		len = pack_Mail(&pilotMail, record, sizeof(record));

		/* write the mail record to the Palm now */
		if (0 < dlp_WriteRecord(gSocket, gDB, 0, 0, 0, record, len, 0)) {
			/* yay, wrote mail to Palm */
			nToPalm++;
			printf("."); /* some feedback - one '.' per delivered mail */
			fflush(stdout);
		} else {
			mm_log("  failure writing to Palm\n", ERROR);
		}
	}
}

/* findMail()
 *
 * find (all) message(s) matching *pilotMail
 * caller is responsible for clearing gSearchResults
 */
listItem *
findMail(struct Mail *pilotMail)
{
	SEARCHPGM search;

	memset(&search, 0, sizeof(SEARCHPGM));

	if (pilotMail->subject)
		search.subject = mkSL(pilotMail->subject);
	if (pilotMail->from)
		search.from = mkSL(pilotMail->from);
	if (pilotMail->to)
		search.to = mkSL(pilotMail->to);
	if (pilotMail->cc)
		search.cc = mkSL(pilotMail->cc);
	if (pilotMail->bcc)
		search.bcc = mkSL(pilotMail->bcc);

	search.senton = (	(pilotMail->date.tm_year - BASEYEAR) << 9) +
						(pilotMail->date.tm_mon << 5) +
						pilotMail->date.tm_mday;

	mail_search_full(gMailStream, gCharSet, &search, NIL);

	if (search.subject)
		delSL(search.subject);
	if (search.from)
		delSL(search.from);
	if (search.to)
		delSL(search.to);
	if (search.cc)
		delSL(search.cc);
	if (search.bcc)
		delSL(search.bcc);

	return gSearchResults;
}

/* markDeleted
 *
 * mark message numbered 'msgno' as Deleted on the server
 */
void
markDeleted(unsigned long msgno)
{
	char seq[5]; /* won't handle >9,999 msg mailboxes */

	sprintf(seq, "%li", msgno);

#ifdef DEBUG
	printf("  marking mail #%s for deletion\n", seq);
#endif

	mail_setflag(gMailStream, seq, "\\DELETED");
}

/* markSeen
 *
 * mark message numbered 'msgno' as Seen on the server
 */
void
markSeen(unsigned long msgno)
{
	char seq[5]; /* won't handle >9,999 msg mailboxes */

	sprintf(seq, "%li", msgno);

#ifdef DEBUG
	printf("  marking mail #%s as seen\n", seq);
#endif

	mail_setflag(gMailStream, seq, "\\SEEN");
}

/* isSeen
 *
 * return true if message number 'msgno' is marked Seen on the server
 * return false otherwise
 */
int
isSeen(unsigned long msgno)
{
	BODY *body;
	MESSAGECACHE *elt;

	mail_fetchstructure(gMailStream, msgno, &body);
 
	if (body) {
		elt = mail_elt(gMailStream, msgno);
		return elt->seen;
	}
	return 0;
}


/* mm_searched: called on every message that matches the conditions set in
 *              mail_search_full (above).
 * 
 */
void
mm_searched (MAILSTREAM *stream, unsigned long number)
{
	push(&gSearchResults, (void *)number);
}

void
mm_login(NETMBX *mb, char *user, char *pwd, long trial)
{
	char *nl=NULL;

#ifdef DEBUG
	printf("acquiring login info\n");
#endif
	if (trial || *gUser == '\0') {
		printf("login: ");
		fgets(gUser, NETMAXUSER, stdin);
		/* get rid of the trailing \n */
		if ((nl=strchr(gUser, '\n')))
			*nl = '\0';
	}

	if (trial || *gPwd == '\0')
		strncpy(gPwd, getpass("Password: "), MAILTMPLEN);

	strncpy(user, gUser, NETMAXUSER);
	strncpy(pwd, gPwd, MAILTMPLEN);
}

void
mm_log (char *string,long errflg)
{
	switch ((short) errflg) {
		case NIL:
			if (gVerbose)
				printf ("-- %s\n", string);
			break;
		case PARSE:
		case WARN:
			printf ("== Warning: %s\n", string);
			break;
		case ERROR:
			printf ("** Error: %s\n", string);
			break;
	}
}

#ifdef DEBUG
void
printMail(const struct Mail *mail)
{
	if (mail->subject) printf("  subject: %s", mail->subject);
	if (mail->from) printf("  from: %s", mail->from);
	if (mail->to) printf("  to: %s", mail->to);
	if (mail->cc) printf("  cc: %s", mail->cc);
	if (mail->bcc) printf("  bcc: %s", mail->bcc);
	if (mail->replyTo) printf("  replyTo: %s", mail->replyTo);
	if (mail->sentTo) printf("  sentTo: %s", mail->sentTo);
}
#endif

char *
getHeader(MAILSTREAM *stream, unsigned long msg_no, const char *headerstr)
{
	STRINGLIST *header_list;
	char *hdr=NULL;

	if (NULL == (header_list = mail_newstringlist())) {
		mm_log("Ran out of memory in getHeader()\n", ERROR);
		return NULL;
	}

	header_list->text.data = (unsigned char *)strdup(headerstr);
	header_list->text.size = strlen((char *)header_list->text.data);

	/* skip 'headerstr: ' */
	hdr = strchr(mail_fetchheader_full(stream,msg_no,header_list,NIL,NIL), ' ');
	if (hdr) {
		hdr++; /* skip leading space */
		/* convert \r\n\r\n to \n\0 */
		hdr[strlen(hdr)-4] = '\n'; 
		hdr[strlen(hdr)-3] = '\0'; 
	}

	free(header_list->text.data);
	return hdr;
}


/***** nothing to see below this point *****/

void mm_notify (MAILSTREAM *stream,char *string,long errflg)
{
	mm_log(string, errflg);
}

long mm_diskerror (MAILSTREAM *stream,long errcode,long serious)
{
	abort();
	return NIL;
}

void mm_fatal (char *string)
{
	mm_log(string, ERROR);
}

void mm_dlog (char *string)
{
	puts(string);
}

void mm_critical (MAILSTREAM *stream)
{
}

void mm_nocritical (MAILSTREAM *stream)
{
}

void smtptest (long debug)
{
}

void mm_exists (MAILSTREAM *stream,unsigned long number)
{
}

void mm_expunged (MAILSTREAM *stream,unsigned long number)
{
}

void mm_flags (MAILSTREAM *stream,unsigned long number)
{
}

void mm_list (MAILSTREAM *stream,int delimiter,char *mailbox,long attributes)
{
}

void mm_lsub (MAILSTREAM *stream,int delimiter,char *mailbox,long attributes)
{
}

void mm_status (MAILSTREAM *stream,char *mailbox,MAILSTATUS *status)
{
}

