/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License. 
 * 
 * The Original Code is pilot-mailsync.
 * 
 * The Initial Developer of the Original Code is Hollis Blanchard
 * <hollis@austin.rr.com>. Portions created by Hollis Blanchard are Copyright
 * (C) 2001 Hollis Blanchard. All Rights Reserved.
 * 
 * Contributor(s):
 * Frederic Devernay <Frederic.Devernay@sophia.inria.fr>
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <pi-source.h>
#include <pi-socket.h>
#include <time.h> /* XXX pilot-link bug: required before pi-mail.h */
#include <pi-mail.h>
#include <pi-dlp.h>

#include "c-client/c-client.h"

#include "cclientcalls.h"
#include "options.h"
#include "mailstore.h"
#include "palmmail.h"
#include "pilot-mailsync.h"
#include "util.h"

/* Some things (gSocket, gDB) HAD to be global so they could be used in
 * the mm_searched callback... and then it made life easier to have the rest
 * global too.
 *
 * Luckily we never have to worry about multiple DB's, sigs, sockets, etc.
 */
struct MailSyncPref gPrefs;
struct MailSignaturePref gSig;
int gSocket=0;
int gDB=0;
int nFromPalm=0;
int nToPalm=0;

/* private */
static void OpenMailDB();
static void CloseMailDB();

/* OpenMailDB()
 *
 * opens the Palm's Mail database, Mail preferences, and signature
 */
void
OpenMailDB(void)
{
	struct PilotUser pi_user;
	int bufsize = 0xffff; /* XXX could probably be smaller */
	unsigned char buffer[bufsize];
	int version;
	int size;

	memset(&gPrefs, 0, sizeof(struct MailSyncPref));
	memset(&gSig, 0, sizeof(struct MailSignaturePref));

	/* ID the Palm (XXX unused but maybe necessary?) */
	dlp_ReadUserInfo(gSocket,&pi_user);

	/* display message on Palm */
	dlp_OpenConduit(gSocket);

	/* Open the Mail database, store access handle in db */
	if (0 > dlp_OpenDB(gSocket, 0, dlpOpenReadWrite, "MailDB", &gDB)) {
		fprintf(stderr, "Unable to open MailDB\n");
		dlp_AddSyncLogEntry(gSocket, "Unable to open MailDB.\n");
		exit(1);
	}

	if (pi_version(gSocket) <= 0x0100) {
		fprintf(stderr, "pilot-mailsync requires a higher version of"
						"the Palm OS than what you've got.\n");
		CloseMailDB();
		exit(1);
	}

	dlp_ReadAppBlock(gSocket, gDB, 0, buffer, 0xffff);

	/* read in Mail prefs (for signature, filter settings, etc) */
	if (0 < dlp_ReadAppPreference(gSocket, makelong("mail"), mailPrefLocal, 1,
									bufsize, buffer, &size, &version) ) {
		unpack_MailSyncPref(&gPrefs, buffer, bufsize);

#ifdef DEBUG
		printf("gPrefs.syncType = %i\n", gPrefs.syncType);
		printf("gPrefs.getHigh = %i\n", gPrefs.getHigh);
		printf("gPrefs.getContaining = %i\n", gPrefs.getContaining);
		printf("gPrefs.truncate = %i\n", gPrefs.truncate);
		printf("gPrefs.filterTo = %s\n", gPrefs.filterTo);
		printf("gPrefs.filterFrom = %s\n", gPrefs.filterFrom);
		printf("gPrefs.filterSubject = %s\n", gPrefs.filterSubject);
#endif

	} else {
		fprintf(stderr, "Couldn't read Mail preferences\n");
		CloseMailDB();
		exit(1);
	}

	/* read in Mail's signature info */
	if (0 < dlp_ReadAppPreference(gSocket, makelong("mail"), mailPrefSig, 1,
									bufsize, buffer, 0, 0) ) {
		unpack_MailSignaturePref(&gSig, buffer, bufsize);
	}

	if (gLocalHack) {
		/* hack for POP3 & MH folks to work around c-client's inability to
		 * perceive new messages in these types of mailboxes
		 */
		int i;
		recordid_t id;
		int size, attr, cat;

		printf("Using POP3/MH hack...\n");

		/* load all Palm emails into memory to compare for dups later */
		initStore();
		for (i=0; 0 < dlp_ReadRecordByIndex(gSocket, gDB, i, buffer, &id,
												&size, &attr, &cat); i++) {
			struct Mail *mail = (struct Mail *)malloc(sizeof(struct Mail));

			/* deleted and archived records already deleted */

			unpack_Mail(mail, buffer, size);
			storeMsg(mail);
		}
	}
}

/* CloseMailDB()
 *
 * closes Palm's mail database
 * writes to Palm hotsync log
 */
void
CloseMailDB(void)
{
	char status[1024];

	dlp_ResetLastSyncPC(gSocket);

	/* Close the database */
	dlp_CloseDB(gSocket, gDB);

	sprintf(status,	"pilot-mailsync " VERSION ":\n"
					"%d message%s sent from handheld.\n"
					"%d message%s delivered to handheld.\n",
					nFromPalm, (nFromPalm == 1) ? "" : "s",
					nToPalm, (nToPalm == 1) ? "" : "s");

	/* print out status info */
	fprintf(stderr, status);	
	/* append to hotsync log on Pilot */
	dlp_AddSyncLogEntry(gSocket, status);

	free_MailSyncPref(&gPrefs);
	free_MailSignaturePref(&gSig);

	if (gLocalHack) {
		/* free our local copy of the Palm MailDB */
		emptyStore();
	}
}

/* cancel()
 *
 * called when user kills process on the desktop to cleanly break connection
 *
 * XXX RETSIGTYPE is apparently coming from pi-config.h - how to use
 * AC_TYPE_SIGNAL?
 */
RETSIGTYPE
cancel(int sig)
{
	char status[1024];

	dlp_CleanUpDatabase(gSocket, gDB); /* delete archived/deleted records */
	dlp_CloseDB(gSocket, gDB); /* close the database */

	sprintf(status,	"\npilot-mailsync " VERSION ":\nuser aborted!\n"
					"%d message%s sent from handheld.\n"
					"%d message%s delivered to handheld.\n",
					nFromPalm, (nFromPalm == 1) ? "" : "s",
					nToPalm, (nToPalm == 1) ? "" : "s");

	fprintf(stderr, status);
	dlp_AddSyncLogEntry(gSocket, status);

	dlp_EndOfSync(gSocket, dlpEndCodeUserCan);
	exit(1);
}


int
main(int argc, char *argv[])
{
	load_options(argc, argv); /* set up globals */
	signal(SIGINT, cancel);

	gSocket = Connect(gDevice); /* attach to Palm */ 
	OpenMailDB(); /* open Palm Mail db and prefs */

	/* only connect to mail source if we'll need something from there */
	if (gPrefs.syncType != mailSyncSend) {
		OpenMail(gMailbox); /* connect to mail source */

		DeleteMails(); /* delete mails from Palm and possibly mail source */
		if (gSyncSeen)
			UpdateSeenFlags(); /* update Seen flags on Palm and mail source */
	 
		MailToPalm(); /* deliver mail to Palm */
	}

	PalmToMail(gFrom); /* send all mail from Palm */

	CloseMail(); /* disconnect from mail source */
	CloseMailDB(); /* close Palm db's */
	Disconnect(gSocket); /* detach from Palm */
	
	return 0;
}
